// VERSION 1.1

//This module decodes a single hexadecimal digit into a pattern for the 7-segment cathodes.
module decode4(HEX_IN, DISP_OUT);
   input [3:0] HEX_IN;
   output [6:0] DISP_OUT;
   reg [6:0]    DISP_OUT;

   always @(HEX_IN)
     begin
        case (HEX_IN)
        //The cathodes are active low, with the most significant bit of DISP_OUT representing
        //cathode g and the least significant bit representing cathode a.
          0:  DISP_OUT = 7'b1000000;  
          1:  DISP_OUT = 7'b1111001;
          2:  DISP_OUT = 7'b0100100;
          3:  DISP_OUT = 7'b0110000;
          4:  DISP_OUT = 7'b0011001;
          5:  DISP_OUT = 7'b0010010;
          6:  DISP_OUT = 7'b0000010;
          7:  DISP_OUT = 7'b1111000;
          8:  DISP_OUT = 7'b0000000;
          9:  DISP_OUT = 7'b0011000;
          10:  DISP_OUT = 7'b0001000;
          11:  DISP_OUT = 7'b0000011;
          12:  DISP_OUT = 7'b0100111;
          13:  DISP_OUT = 7'b0100001;
          14:  DISP_OUT = 7'b0000110;
          15:  DISP_OUT = 7'b0001110;
          default:  DISP_OUT = 7'b1111111;  //We will never reach the default case
        endcase
     end

endmodule



//This module is a simple 4-bit counter, used to select the active anode of the 7-segment display
//and it's corresponding decoded digit. 
module counter (C, Q);
   input C;             //Clock input
   output [1:0] Q;      
   reg [1:0]    Q;

   initial begin
      Q = 0;            //Initialize the output reg to 0, so that incrementing Q is well defined.
   end

   always @(posedge C)    
     begin      
        if (Q >= 3 || Q < 0)  //Explicitly handle all edge/error cases by reseting the counter to 0.      
          Q = 0;      
        else                  //On each positive clock edge, the counter is incremented.
          Q = Q + 1;      
     end  
   
endmodule



//This module is a parameterized, counter-based clock-divider.
module clkdiv (CIN, COUT);
   parameter n = 65535;  // The 7-segment display uses a very slow clock, so the default parameter is large.
   input     CIN;        // Clock input
   output    COUT;       // Output is the input clock divided by n.
   reg       COUT;
   reg [31:0] Q;         // Internal reg to maintain the state of the counter.

   initial begin
      Q = 0;
      COUT = 0;
   end

   always @(posedge CIN)     
     begin      
        if (Q >= n - 1 || Q < 0)  // We have incremented the counter n times 
          begin                   // so we have had n positive edges of the input clock
             Q = 0;               // Reset the counter
             COUT = ~COUT;        // Toggle the output clock.
          end      
        else        
          Q = Q + 1;              // We have not yet had n input clock pos. edges, so keep counting.
     end  
   
endmodule




//This module controls the 3 memory-mapped IO devices on the dio4 extension board.
module di04(SWIN, BTNIN, COUT, AOUT, LDOUT, reg_clk, rst, we, gwe, SEV_SEG_CLK, 
	    lights_data, data_select, read_data,
	    /* debug signals */
	    debug_mode_select, debug_pc_out, debug_imem_out, 
	    debug_reg_data_in, debug_reg_we, debug_dmem_addr,
	    debug_dmem_out, debug_dmem_in, debug_dmem_we);

   /* reg_clk is the 200 Mhz system clock, used to clock the internal registers.  
    * SEV_SEG_CLK is a 32 MHz clock, used to alternate the cathodes and anodes of the 7-seg display.
    * data_select controls which device is active
    *	   001 => LEDs
    *	   010 => seven-segment display
    *	   100 => buttons and switches	
    * SWIN is the state of the 8 expansion board switches
    * BTNIN is the state of the 5 expansion board buttons
    * lights_data is the input data to be written to the LEDs or seven-segment display
    * COUT are the 7 outputs to the cathodes of the seven-segment display
    * AOUT are the 4 outputs to the anodes of the seven-segment display
    * LDOUT are the 8 outputs to the expansion board LEDs
    * LEDG is a controller for the expansion board LEDs.  When LEDG is 1, the LEDs are enabled for writing.
    * read_data is the output of the currently active device (see data_select) for memory reads
    */

   input        reg_clk, rst, we, gwe, SEV_SEG_CLK;			 
   input [2:0] 	data_select;
   input [7:0] 	SWIN;
   input [4:0] 	BTNIN;
   input [15:0] lights_data;

   input [2:0] 	debug_mode_select;
   input        debug_reg_we, debug_dmem_we;
   input [15:0] debug_pc_out, debug_imem_out, debug_reg_data_in;
   input [15:0] debug_dmem_addr, debug_dmem_out, debug_dmem_in;

   output [6:0] COUT;
   output [3:0] AOUT;
   output [7:0] LDOUT;
   output [15:0] read_data;

   // We should write to the LEDs if we are performing a write (we == 1) and the LEDs are the
   // active device (data_select == 001).  Note that data_select is one-hot.

   wire          led_we;                     // Write-enable signal for the LED register.
   and(led_we, we, data_select[0]);

   // We should write to the seven_segment display if we are performing a write (we == 1) 
   // and the 7-seg is the active device (data_select == 010).

   wire          seven_seg_we;              // Write-enable signal for the 7-seg register.
   and(seven_seg_we, we, data_select[1]);


   /* Since the 7-segment display shares the 7 cathodes for all 4 digits (anodes), to create 
    * an output with multiple digits, we must rapdily alternate between the 4 digits,
    * changing the cathodes to the appropriate value for that digit.  This alternation must
    * be faster than 60 Hz, so that we cannot see the alternation, but it must also be sufficiently
    * slow.  If the alternation is too fast, then all active cathodes from all 4 digits will appear 
    * to be lit at all times. slow_clk is a obtained by slowing down the input clock to a
    * reasonable speed for the seven-segment display.
   */  
   wire          slow_clk;
   clkdiv #(65535) sclock( .CIN(SEV_SEG_CLK), .COUT(slow_clk) );

   // A register is used to synchronize the seven-segment display with the rest of the memory system.
   wire [15:0]   seven_seg_data;
   Nbit_reg #(16, 0) seven_seg_reg(.clk(reg_clk), 
                                       .we(seven_seg_we), 
                                       .gwe(gwe),
                                       .rst(rst), 
                                       .in(lights_data), 
                                       .out(seven_seg_data));

   // A register is used to synchronize the LEDs with the rest of the memory system.
   wire [7:0]    led_data;
   Nbit_reg #(8, 0) led_reg(.clk(reg_clk), 
				.we(led_we), 
				.gwe(gwe),
				.rst(rst), 
				.in(lights_data[7:0]), 
				.out(led_data));

   // select is used to choose the active 7-seg anode and its corresponding data.
   // Since there are 4 anodes, this is done with a counter based on the slow_clk
   wire [1:0]    select, select_;
   counter count( .C(slow_clk), 
		  .Q(select) );

   not(select_[0], select[0]);
   not(select_[1], select[1]);

   // Each digit represents the proper cathode settings for 4-bits of the seven-segment data
   wire [6:0] 	 first_digit, second_digit, third_digit, fourth_digit;

   wire [15:0] 	 seven_seg_data_out;

   mux_Nbit_8to1 #(16) sev_seg_data_mux(.a(seven_seg_data),
					.b(debug_pc_out),
					.c(debug_imem_out),
					.d(debug_reg_data_in),
					.e( { {3{1'b0}}, rst, {4{1'b0}}, {3{1'b0}}, debug_reg_we, {3{1'b0}}, debug_dmem_we } ),
					.f(debug_dmem_addr),
					.g(debug_dmem_out),
					.h(debug_dmem_in),
					
					.sel(debug_mode_select),
					.out(seven_seg_data_out));

   decode4 high4 (seven_seg_data_out[15:12], fourth_digit);
   decode4 third4 (seven_seg_data_out[11:8], third_digit);
   decode4 second4 (seven_seg_data_out[7:4], second_digit);
   decode4 low4 (seven_seg_data_out[3:0], first_digit);

   nand(AOUT[3], select_[1], select_[0]);
   nand(AOUT[2], select_[1], select[0]);
   nand(AOUT[1], select[1], select_[0]);
   nand(AOUT[0], select[1], select[0]);

   mux_Nbit_4to1 #(7) cathode_mux(.a(first_digit), 
                                  .b(second_digit), 
                                  .c(third_digit), 
                                  .d(fourth_digit), 
                                  .sel(select), 
                                  .out(COUT) );


   // The led_data can be sent unmodified to the LEDs.
   assign        LDOUT = led_data;


   //When performing a memory read from the devices, choose the proper device based on data_select.
   assign        read_data = data_select == 3'b001 ? led_data :
                             data_select == 3'b010 ? seven_seg_data :
                             data_select == 3'b100 ? {3'b000, BTNIN[4:0], SWIN[7:0]}:
                             16'h0000;    //This is an error -- return 0x0000


endmodule
